package egs.wolves;

import util.Rand;

import java.util.*;
import static egs.wolves.Move.*;
import static egs.wolves.Grid.*;

import java.io.*;

abstract class Rabbit {
  private final String name;
  Rabbit(String nm){
    name = nm;
  }
  String name() { return name; }

  abstract Move pickMove(Grid w);




  /**
   * A rabbit that stays as far as it can from both wolves.
   * It does this by seeing who it is closest to and choosing the
   * move that moves us furthest from the rabbit.
   */
  static Rabbit skittish(){
    return SKITTISH;
  }

  private static final Move[] ALL_MOVES = Move.values();

  private static final Rabbit SKITTISH = new Rabbit("skittish") {
    Move pickMove(Grid w) {
      Move bestMv = null;
      int bestDist = -1;
      for (int i = 0; i < ALL_MOVES.length; i++) {
        Move m = ALL_MOVES[i];
        int d = w.moveRabbit(m).closestWolfDist();
        if (d > bestDist) {
          bestMv = m;
          bestDist = d;
        }
      }
      return bestMv;
    }
  };

  /**
   * Like skittish, but randomly choose between optimal moves.
   */
  static Rabbit skittishRandom(long seed){
    final Random rand = new Random(seed);
    return new Rabbit("randomized-skittish("+seed+")") {
      Move pickMove(Grid g) {
        List<Move> allOpt = new ArrayList<Move>();
        int bestDist = -1;
        for (int i = 0; i < ALL_MOVES.length; i++) {
          Move m = ALL_MOVES[i];
          int d = g.moveRabbit(m).closestWolfDist();
          if (d > bestDist) {
            bestDist = d;
            allOpt.clear();
            allOpt.add(m);
          } else if (d == bestDist) {
            allOpt.add(m);
          }
        }
        return Rand.pickElement(rand, allOpt);
      }
    };
  }

  /**
   * A rabbit that chooses its moves uniformly randomly
   * irrespective of its or the wolves' positions
   */
  static Rabbit random(long seed){
    final Move[] ALL_MOVES = Move.values();
    final Random gen = new Random(seed);
    return new Rabbit("random("+seed+")") {
      Move pickMove(Grid g){
        return Rand.pickElement(gen, g.validMoves(2));
//        return Rand.pickElement(gen, ALL_MOVES);
      }
    };
  }

  static Rabbit random(){
    return random(System.currentTimeMillis());
  }


  static Rabbit interactive(){
    final BufferedReader br = new BufferedReader(new InputStreamReader(System.in));
    final Rabbit skittish = skittish();
    final Rabbit random = random(System.currentTimeMillis());
    return new Rabbit("interactive") {
      Move pickMove(Grid w) {
        while (true) {
          System.out.println("CHOOSE A MOVE:");
          String line = null;
          try {
            line = br.readLine();
          } catch (IOException iox){ iox.printStackTrace(); }
          if (line == null) {
            throw new RuntimeException("EOF reached");
          }
          line = line.trim();

          if (line.equals("null")) {
            return null;

          } else if (line.equalsIgnoreCase("L")) {
            return Move.LEFT;

          } else if (line.equalsIgnoreCase("R")) {
            return Move.RIGHT;

          } else if (line.equalsIgnoreCase("U")) {
            return Move.UP;

          } else if (line.equalsIgnoreCase("D")) {
            return Move.DOWN;

          } else if (line.equalsIgnoreCase("S")) {
            return Move.STAY;

          } else if (line.equalsIgnoreCase("X")) {
            System.out.println("Exiting");
            System.exit(0);

          } else if (line.equals("")) {
            Move m = random.pickMove(w);
            System.out.println("Randomly chose "+m);
            return m;

          } else if (line.equalsIgnoreCase("K")) {
            Move m = skittish.pickMove(w);
            System.out.println("Skittish choice is "+m);
            return m;

          } else {
            System.err.println("Unrecognized command, choose from:");
            System.err.println("(L,R,U,D,S,X,<empty>,K) [(L)eft, (R)ight, (U)p, "+
                                  "(D)own, (S)tay, e(X)it, <empty string> = random, s(K)ittish");
          }
        }
      }
    };
  }
}
